/*
 * Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_MRU_H__
#define __VIRTNET_DPA_MRU_H__

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#define	_MRU_HEAD(name, idx_type)					\
struct name {								\
	idx_type first;		/* first element */			\
	idx_type last;		/* last element */			\
	idx_type num;		/* num of element */			\
}
#define	MRU_HEAD(name, idx_type) _MRU_HEAD(name, idx_type)

#define	_MRU_ENTRY(idx_type)						\
struct {								\
	idx_type prev;		/* previous element */			\
	idx_type next;		/* next element */			\
}
#define	MRU_ENTRY(idx_type) _MRU_ENTRY(idx_type)

#define	MRU_INIT(head)							\
do {									\
	(head)->first = 0;						\
	(head)->last = 0;						\
	(head)->num = 0;						\
} while (0)

/* Must be called with lock */
#define	MRU_INSERT_HEAD(head, array, idx, entry_type, field)		\
do {									\
	struct entry_type *elm_ = &(array)[(idx)];			\
	if ((head)->num != 0) {						\
		elm_->field.next = (head)->first;			\
		(array)[(head)->first].field.prev = (idx);		\
	} else								\
		(head)->last = (idx);					\
									\
	(head)->first = (idx);						\
	(head)->num++;							\
} while (0)

/* Must be called with lock */
#define	MRU_REMOVE(head, array, idx, entry_type, field)			\
do {									\
	struct entry_type *elm_ = &(array)[(idx)];			\
	if ((idx) != (head)->last)					\
		(array)[elm_->field.next].field.prev =			\
			elm_->field.prev;				\
	else								\
		(head)->last = elm_->field.prev;			\
									\
	if ((idx) != (head)->first)					\
		(array)[elm_->field.prev].field.next =			\
			elm_->field.next;				\
	else								\
		(head)->first = elm_->field.next;			\
									\
	elm_->field.prev = 0;						\
	elm_->field.next = 0;						\
	(head)->num--;							\
} while (0)

/* Must be called with lock */
#define	MRU_MOVE_TO_HEAD(head, array, idx, entry_type, field)		\
do {									\
	if ((idx) == (head)->first)					\
		break;							\
	MRU_REMOVE(head, array, idx, entry_type, field);		\
	MRU_INSERT_HEAD(head, array, idx, entry_type, field);		\
} while (0)

/* Must be called with lock */
#define	MRU_FOREACH(var, head, array, i, field)				\
	for ((var) = &((array)[(head)->first]), (i) = 0;		\
		(i)++ < (head)->num;					\
		(var) = &((array)[(var)->field.next]))			\

/* Must be called with lock */
#define	MRU_FOREACH_REVERSE(var, head, array, i, field)			\
	for ((var) = &((array)[(head)->last]), (i) = 0;			\
		(i)++ < (head)->num;					\
		(var) = &((array)[(var)->field.prev]))			\

#define	MRU_LAST(head, array) (&((array)[(head)->last]))

#define	MRU_FIRST(head, array) (&((array)[(head)->first]))

/* Must be called with lock */
#define	MRU_MUTIPLE_REMOVE(head, array, entry, del_num, field)		\
do {									\
	if ((entry) == MRU_FIRST((head), (array)))			\
		MRU_INIT((head));					\
	else {								\
		(array)[(entry)->field.prev].mru_entry.next = 0;	\
		(head)->last = (entry)->field.prev;			\
		(head)->num -= (del_num);				\
	}								\
} while (0)

#ifdef __cplusplus
}
#endif

#endif
